@@ -1,28 +1,58 @@
-# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
 
-"""
-
+r"""
 .. dialect:: oracle+cx_oracle
     :name: cx-Oracle
     :dbapi: cx_oracle
-    :connectstring: oracle+cx_oracle://user:pass@host:port/dbname\
-[?key=value&key=value...]
+    :connectstring: oracle+cx_oracle://user:pass@host:port/dbname[?key=value&key=value...]
     :url: http://cx-oracle.sourceforge.net/
 
 Additional Connect Arguments
 ----------------------------
 
-When connecting with ``dbname`` present, the host, port, and dbname tokens are
-converted to a TNS name using
-the cx_oracle ``makedsn()`` function.  Otherwise, the host token is taken
-directly as a TNS name.
+When connecting with the ``dbname`` URL token present, the ``hostname``,
+``port``, and ``dbname`` tokens are converted to a TNS name using the
+``cx_Oracle.makedsn()`` function. The URL below::
 
-Additional arguments which may be specified either as query string arguments
-on the URL, or as keyword arguments to :func:`.create_engine()` are:
+    e = create_engine("oracle+cx_oracle://user:pass@hostname/dbname")
+
+Will be used to create the DSN as follows::
+
+    >>> import cx_Oracle
+    >>> cx_Oracle.makedsn("hostname", 1521, sid="dbname")
+    '(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST=hostname)(PORT=1521))(CONNECT_DATA=(SID=dbname)))'
+
+The ``service_name`` parameter, also consumed by ``cx_Oracle.makedsn()``, may
+be specified in the URL query string, e.g. ``?service_name=my_service``.
+
+If ``dbname`` is not present, then the value of ``hostname`` in the
+URL is used directly as the DSN passed to ``cx_Oracle.connect()``.
+
+Additional connection arguments may be sent to the ``cx_Oracle.connect()``
+function using the :paramref:`.create_engine.connect_args` dictionary.
+Any cx_Oracle parameter value and/or constant may be passed, such as::
+
+    import cx_Oracle
+    e = create_engine(
+        "oracle+cx_oracle://user:pass@dsn",
+        connect_args={
+            "mode": cx_Oracle.SYSDBA,
+            "events": True
+        }
+    )
+
+There are also options that are consumed by the SQLAlchemy cx_oracle dialect
+itself.  These options are always passed directly to :func:`.create_engine`,
+such as::
+
+    e = create_engine(
+        "oracle+cx_oracle://user:pass@dsn", coerce_to_unicode=False)
+
+The parameters accepted by the cx_oracle dialect are as follows:
 
 * ``arraysize`` - set the cx_oracle.arraysize value on cursors, defaulted
   to 50.  This setting is significant with cx_Oracle as the contents of LOB
@@ -35,21 +65,10 @@ on the URL, or as keyword arguments to :func:`.create_engine()` are:
 
 * ``coerce_to_decimal`` - see :ref:`cx_oracle_numeric` for detail.
 
-* ``mode`` - This is given the string value of SYSDBA or SYSOPER, or
-  alternatively an integer value.  This value is only available as a URL query
-  string argument.
-
-* ``threaded`` - enable multithreaded access to cx_oracle connections.
-  Defaults to ``True``.  Note that this is the opposite default of the
-  cx_Oracle DBAPI itself.
-
-* ``service_name`` - An option to use connection string (DSN) with
-  ``SERVICE_NAME`` instead of ``SID``. It can't be passed when a ``database``
-  part is given.
-  E.g. ``oracle+cx_oracle://scott:tiger@host:1521/?service_name=hr``
-  is a valid url. This value is only available as a URL query string argument.
-
-  .. versionadded:: 1.0.0
+* ``threaded`` - this parameter is passed as the value of "threaded" to
+  ``cx_Oracle.connect()`` and defaults to True, which is the opposite of
+  cx_Oracle's default.   This parameter is deprecated and will default to
+  ``False`` in version 1.3 of SQLAlchemy.
 
 .. _cx_oracle_unicode:
 
@@ -93,6 +112,107 @@ Python 2:
   The above approach will add significant latency to result-set fetches
   of plain string values.
 
+Sending String Values as Unicode or Non-Unicode
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+
+As of SQLAlchemy 1.2.2, the cx_Oracle dialect unconditionally calls
+``setinputsizes()`` for bound values that are passed as Python unicode objects.
+In Python 3, all string values are Unicode; for cx_Oracle, this corresponds to
+``cx_Oracle.NCHAR`` being passed to ``setinputsizes()`` for that parameter.
+In some edge cases, such as passing format specifiers for
+the ``trunc()`` function, Oracle does not accept these as NCHAR::
+
+    from sqlalchemy import func
+
+    conn.execute(
+        func.trunc(func.sysdate(), 'dd')
+    )
+
+In these cases, an error as follows may be raised::
+
+    ORA-01899: bad precision specifier
+
+When this error is encountered, it may be necessary to pass the string value
+with an explicit non-unicode type::
+
+    from sqlalchemy import func
+    from sqlalchemy import literal
+    from sqlalchemy import String
+
+    conn.execute(
+        func.trunc(func.sysdate(), literal('dd', String))
+    )
+
+For full control over this ``setinputsizes()`` behavior, see the section
+:ref:`cx_oracle_setinputsizes`
+
+.. _cx_oracle_setinputsizes:
+
+Fine grained control over cx_Oracle data binding performance with setinputsizes
+-------------------------------------------------------------------------------
+
+The cx_Oracle DBAPI has a deep and fundamental reliance upon the usage of the
+DBAPI ``setinputsizes()`` call.   The purpose of this call is to establish the
+datatypes that are bound to a SQL statement for Python values being passed as
+parameters.  While virtually no other DBAPI assigns any use to the
+``setinputsizes()`` call, the cx_Oracle DBAPI relies upon it heavily in its
+interactions with the Oracle client interface, and in some scenarios it is  not
+possible for SQLAlchemy to know exactly how data should be bound, as some
+settings can cause profoundly different performance characteristics, while
+altering the type coercion behavior at the same time.
+
+Users of the cx_Oracle dialect are **strongly encouraged** to read through
+cx_Oracle's list of built-in datatype symbols at
+http://cx-oracle.readthedocs.io/en/latest/module.html#types.
+Note that in some cases, signficant performance degradation can occur when
+using these types vs. not, in particular when specifying ``cx_Oracle.CLOB``.
+
+On the SQLAlchemy side, the :meth:`.DialectEvents.do_setinputsizes` event can
+be used both for runtime visibliity (e.g. logging) of the setinputsizes step as
+well as to fully control how ``setinputsizes()`` is used on a per-statement
+basis.
+
+.. versionadded:: 1.2.9 Added :meth:`.DialectEvents.setinputsizes`
+
+
+Example 1 - logging all setinputsizes calls
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+
+The following example illustrates how to log the intermediary values from a
+SQLAlchemy perspective before they are converted to the raw ``setinputsizes()``
+parameter dictionary.  The keys of the dictionary are :class:`.BindParameter`
+objects which have a ``.key`` and a ``.type`` attribute::
+
+    from sqlalchemy import create_engine, event
+
+    engine = create_engine("oracle+cx_oracle://scott:tiger@host/xe")
+
+    @event.listens_for(engine, "do_setinputsizes")
+    def _log_setinputsizes(inputsizes, cursor, statement, parameters, context):
+        for bindparam, dbapitype in inputsizes.items():
+                log.info(
+                    "Bound parameter name: %s  SQLAlchemy type: %r  "
+                    "DBAPI object: %s",
+                    bindparam.key, bindparam.type, dbapitype)
+
+Example 2 - remove all bindings to CLOB
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+
+The ``CLOB`` datatype in cx_Oracle incurs a significant performance overhead,
+however is set by default for the ``Text`` type within the SQLAlchemy 1.2
+series.   This setting can be modified as follows::
+
+    from sqlalchemy import create_engine, event
+    from cx_Oracle import CLOB
+
+    engine = create_engine("oracle+cx_oracle://scott:tiger@host/xe")
+
+    @event.listens_for(engine, "do_setinputsizes")
+    def _remove_clob(inputsizes, cursor, statement, parameters, context):
+        for bindparam, dbapitype in list(inputsizes.items()):
+            if dbapitype is CLOB:
+                del inputsizes[bindparam]
+
 
 .. _cx_oracle_returning:
 
@@ -177,35 +297,42 @@ SQLAlchemy type (or a subclass of such).
    reworked to take advantage of newer cx_Oracle features as well
    as better integration of outputtypehandlers.
 
-"""
+"""  # noqa
 
 from __future__ import absolute_import
 
-from .base import OracleCompiler, OracleDialect, OracleExecutionContext
-from . import base as oracle
-from ...engine import result as _result
-from sqlalchemy import types as sqltypes, util, exc, processors
-import random
 import collections
 import decimal
+import random
 import re
-import time
+
+from . import base as oracle
+from .base import OracleCompiler
+from .base import OracleDialect
+from .base import OracleExecutionContext
+from ... import exc
+from ... import processors
+from ... import types as sqltypes
+from ... import util
+from ...engine import result as _result
 
 
 class _OracleInteger(sqltypes.Integer):
+    def get_dbapi_type(self, dbapi):
+        # see https://github.com/oracle/python-cx_Oracle/issues/
+        # 208#issuecomment-409715955
+        return int
+
     def _cx_oracle_var(self, dialect, cursor):
         cx_Oracle = dialect.dbapi
         return cursor.var(
-            cx_Oracle.STRING,
-            255,
-            arraysize=cursor.arraysize,
-            outconverter=int
+            cx_Oracle.STRING, 255, arraysize=cursor.arraysize, outconverter=int
         )
 
     def _cx_oracle_outputtypehandler(self, dialect):
-        def handler(cursor, name,
-                    default_type, size, precision, scale):
+        def handler(cursor, name, default_type, size, precision, scale):
             return self._cx_oracle_var(dialect, cursor)
+
         return handler
 
 
@@ -217,7 +344,8 @@ class _OracleNumeric(sqltypes.Numeric):
             return None
         elif self.asdecimal:
             processor = processors.to_decimal_processor_factory(
-                decimal.Decimal, self._effective_decimal_return_scale)
+                decimal.Decimal, self._effective_decimal_return_scale
+            )
 
             def process(value):
                 if isinstance(value, (int, float)):
@@ -226,6 +354,7 @@ class _OracleNumeric(sqltypes.Numeric):
                     return float(value)
                 else:
                     return value
+
             return process
         else:
             return processors.to_float
@@ -283,20 +412,32 @@ class _OracleNumeric(sqltypes.Numeric):
                         type_ = cx_Oracle.NATIVE_FLOAT
 
             return cursor.var(
-                type_, 255,
+                type_,
+                255,
                 arraysize=cursor.arraysize,
-                outconverter=outconverter
+                outconverter=outconverter,
             )
 
         return handler
 
 
+class _OracleBinaryFloat(_OracleNumeric):
+    def get_dbapi_type(self, dbapi):
+        return dbapi.NATIVE_FLOAT
+
+
+class _OracleBINARY_FLOAT(_OracleBinaryFloat, oracle.BINARY_FLOAT):
+    pass
+
+
+class _OracleBINARY_DOUBLE(_OracleBinaryFloat, oracle.BINARY_DOUBLE):
+    pass
+
+
 class _OracleNUMBER(_OracleNumeric):
     is_number = True
 
 
-
-
 class _OracleDate(sqltypes.Date):
     def bind_processor(self, dialect):
         return None
@@ -307,6 +448,7 @@ class _OracleDate(sqltypes.Date):
                 return value.date()
             else:
                 return value
+
         return process
 
 
@@ -341,6 +483,7 @@ class _OracleEnum(sqltypes.Enum):
         def process(value):
             raw_str = enum_proc(value)
             return raw_str
+
         return process
 
 
@@ -361,7 +504,8 @@ class _OracleBinary(sqltypes.LargeBinary):
             return None
         else:
             return super(_OracleBinary, self).result_processor(
-                dialect, coltype)
+                dialect, coltype
+            )
 
 
 class _OracleInterval(oracle.INTERVAL):
@@ -382,14 +526,18 @@ class OracleCompiler_cx_oracle(OracleCompiler):
     _oracle_cx_sql_compiler = True
 
     def bindparam_string(self, name, **kw):
-        quote = getattr(name, 'quote', None)
-        if quote is True or quote is not False and \
-                self.preparer._bindparam_requires_quotes(name):
-            if kw.get('expanding', False):
+        quote = getattr(name, "quote", None)
+        if (
+            quote is True
+            or quote is not False
+            and self.preparer._bindparam_requires_quotes(name)
+        ):
+            if kw.get("expanding", False):
                 raise exc.CompileError(
                     "Can't use expanding feature with parameter name "
                     "%r on Oracle; it requires quoting which is not supported "
-                    "in this context." % name)
+                    "in this context." % name
+                )
             quoted_name = '"%s"' % name
             self._quoted_bind_names[name] = quoted_name
             return OracleCompiler.bindparam_string(self, quoted_name, **kw)
@@ -416,21 +564,22 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
                 if bindparam.isoutparam:
                     name = self.compiled.bind_names[bindparam]
                     type_impl = bindparam.type.dialect_impl(self.dialect)
-                    if hasattr(type_impl, '_cx_oracle_var'):
+                    if hasattr(type_impl, "_cx_oracle_var"):
                         self.out_parameters[name] = type_impl._cx_oracle_var(
-                            self.dialect, self.cursor)
+                            self.dialect, self.cursor
+                        )
                     else:
                         dbtype = type_impl.get_dbapi_type(self.dialect.dbapi)
                         if dbtype is None:
                             raise exc.InvalidRequestError(
                                 "Cannot create out parameter for parameter "
                                 "%r - its type %r is not supported by"
-                                " cx_oracle" %
-                                (bindparam.key, bindparam.type)
+                                " cx_oracle" % (bindparam.key, bindparam.type)
                             )
                         self.out_parameters[name] = self.cursor.var(dbtype)
-                    self.parameters[0][quoted_bind_names.get(name, name)] = \
-                        self.out_parameters[name]
+                    self.parameters[0][
+                        quoted_bind_names.get(name, name)
+                    ] = self.out_parameters[name]
 
     def _generate_cursor_outputtype_handler(self):
         output_handlers = {}
@@ -438,8 +587,9 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
         for (keyname, name, objects, type_) in self.compiled._result_columns:
             handler = type_._cached_custom_processor(
                 self.dialect,
-                'cx_oracle_outputtypehandler',
-                self._get_cx_oracle_type_handler)
+                "cx_oracle_outputtypehandler",
+                self._get_cx_oracle_type_handler,
+            )
 
             if handler:
                 denormalized_name = self.dialect.denormalize_name(keyname)
@@ -448,16 +598,18 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
         if output_handlers:
             default_handler = self._dbapi_connection.outputtypehandler
 
-            def output_type_handler(cursor, name, default_type,
-                                    size, precision, scale):
+            def output_type_handler(
+                cursor, name, default_type, size, precision, scale
+            ):
                 if name in output_handlers:
                     return output_handlers[name](
-                        cursor, name,
-                        default_type, size, precision, scale)
+                        cursor, name, default_type, size, precision, scale
+                    )
                 else:
                     return default_handler(
                         cursor, name, default_type, size, precision, scale
                     )
+
             self.cursor.outputtypehandler = output_type_handler
 
     def _get_cx_oracle_type_handler(self, impl):
@@ -477,7 +629,7 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
 
         self.set_input_sizes(
             self.compiled._quoted_bind_names,
-            include_types=self.dialect._include_setinputsizes
+            include_types=self.dialect._include_setinputsizes,
         )
 
         self._handle_out_parameters()
@@ -494,7 +646,7 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
     def get_result_proxy(self):
         if self.out_parameters and self.compiled.returning:
             returning_params = [
-                self.out_parameters["ret_%d" % i].getvalue()
+                self.dialect._returningval(self.out_parameters["ret_%d" % i])
                 for i in range(len(self.out_parameters))
             ]
             return ReturningResultProxy(self, returning_params)
@@ -502,29 +654,35 @@ class OracleExecutionContext_cx_oracle(OracleExecutionContext):
         result = _result.ResultProxy(self)
 
         if self.out_parameters:
-            if self.compiled_parameters is not None and \
-                    len(self.compiled_parameters) == 1:
+            if (
+                self.compiled_parameters is not None
+                and len(self.compiled_parameters) == 1
+            ):
                 result.out_parameters = out_parameters = {}
 
                 for bind, name in self.compiled.bind_names.items():
                     if name in self.out_parameters:
-                        type = bind.type
-                        impl_type = type.dialect_impl(self.dialect)
+                        type_ = bind.type
+                        impl_type = type_.dialect_impl(self.dialect)
                         dbapi_type = impl_type.get_dbapi_type(
-                            self.dialect.dbapi)
-                        result_processor = impl_type.\
-                            result_processor(self.dialect,
-                                             dbapi_type)
+                            self.dialect.dbapi
+                        )
+                        result_processor = impl_type.result_processor(
+                            self.dialect, dbapi_type
+                        )
                         if result_processor is not None:
-                            out_parameters[name] = \
-                                result_processor(
-                                    self.out_parameters[name].getvalue())
+                            out_parameters[name] = result_processor(
+                                self.dialect._paramval(
+                                    self.out_parameters[name]
+                                )
+                            )
                         else:
-                            out_parameters[name] = self.out_parameters[
-                                name].getvalue()
+                            out_parameters[name] = self.dialect._paramval(
+                                self.out_parameters[name]
+                            )
             else:
                 result.out_parameters = dict(
-                                            (k, v.getvalue())
+                    (k, self._dialect._paramval(v))
                     for k, v in self.out_parameters.items()
                 )
 
@@ -542,14 +700,11 @@ class ReturningResultProxy(_result.FullyBufferedResultProxy):
     def _cursor_description(self):
         returning = self.context.compiled.returning
         return [
-            (getattr(col, 'name', col.anon_label), None)
-            for col in returning
+            (getattr(col, "name", col.anon_label), None) for col in returning
         ]
 
     def _buffer_rows(self):
-        return collections.deque(
-            [tuple(self._returning_params)]
-        )
+        return collections.deque([tuple(self._returning_params)])
 
 
 class OracleDialect_cx_oracle(OracleDialect):
@@ -564,12 +719,13 @@ class OracleDialect_cx_oracle(OracleDialect):
 
     driver = "cx_oracle"
 
-    colspecs = colspecs = {
+    colspecs = {
         sqltypes.Numeric: _OracleNumeric,
         sqltypes.Float: _OracleNumeric,
+        oracle.BINARY_FLOAT: _OracleBINARY_FLOAT,
+        oracle.BINARY_DOUBLE: _OracleBINARY_DOUBLE,
         sqltypes.Integer: _OracleInteger,
         oracle.NUMBER: _OracleNUMBER,
-
         sqltypes.Date: _OracleDate,
         sqltypes.LargeBinary: _OracleBinary,
         sqltypes.Boolean: oracle._OracleBoolean,
@@ -580,7 +736,6 @@ class OracleDialect_cx_oracle(OracleDialect):
         sqltypes.UnicodeText: _OracleUnicodeText,
         sqltypes.CHAR: _OracleChar,
         sqltypes.Enum: _OracleEnum,
-
         oracle.LONG: _OracleLong,
         oracle.RAW: _OracleRaw,
         sqltypes.Unicode: _OracleNVarChar,
@@ -590,13 +745,15 @@ class OracleDialect_cx_oracle(OracleDialect):
 
     execute_sequence_format = list
 
-    def __init__(self,
-                 auto_convert_lobs=True,
-                 threaded=True,
-                 coerce_to_unicode=False,
-                 coerce_to_decimal=True,
-                 arraysize=50,
-                 **kwargs):
+    def __init__(
+        self,
+        auto_convert_lobs=True,
+        threaded=True,
+        coerce_to_unicode=False,
+        coerce_to_decimal=True,
+        arraysize=50,
+        **kwargs
+    ):
 
         self._pop_deprecated_kwargs(kwargs)
 
@@ -616,28 +773,56 @@ class OracleDialect_cx_oracle(OracleDialect):
             self.cx_oracle_ver = self._parse_cx_oracle_ver(cx_Oracle.version)
             if self.cx_oracle_ver < (5, 2) and self.cx_oracle_ver > (0, 0, 0):
                 raise exc.InvalidRequestError(
-                    "cx_Oracle version 5.2 and above are supported")
+                    "cx_Oracle version 5.2 and above are supported"
+                )
 
             self._has_native_int = hasattr(cx_Oracle, "NATIVE_INT")
 
             self._include_setinputsizes = {
-                cx_Oracle.NCLOB, cx_Oracle.CLOB, cx_Oracle.LOB,
-                cx_Oracle.NCHAR, cx_Oracle.FIXED_NCHAR,
-                cx_Oracle.BLOB, cx_Oracle.FIXED_CHAR, cx_Oracle.TIMESTAMP
+                cx_Oracle.NCLOB,
+                cx_Oracle.CLOB,
+                cx_Oracle.LOB,
+                cx_Oracle.NCHAR,
+                cx_Oracle.FIXED_NCHAR,
+                cx_Oracle.BLOB,
+                cx_Oracle.FIXED_CHAR,
+                cx_Oracle.TIMESTAMP,
+                _OracleInteger,
+                _OracleBINARY_FLOAT,
+                _OracleBINARY_DOUBLE,
             }
 
-        self._is_cx_oracle_6 = self.cx_oracle_ver >= (6, )
+            self._paramval = lambda value: value.getvalue()
+
+            # https://github.com/oracle/python-cx_Oracle/issues/176#issuecomment-386821291
+            # https://github.com/oracle/python-cx_Oracle/issues/224
+            self._values_are_lists = self.cx_oracle_ver >= (6, 3)
+            if self._values_are_lists:
+                cx_Oracle.__future__.dml_ret_array_val = True
+
+                def _returningval(value):
+                    try:
+                        return value.values[0][0]
+                    except IndexError:
+                        return None
+
+                self._returningval = _returningval
+            else:
+                self._returningval = self._paramval
+
+        self._is_cx_oracle_6 = self.cx_oracle_ver >= (6,)
 
     def _pop_deprecated_kwargs(self, kwargs):
-        auto_setinputsizes = kwargs.pop('auto_setinputsizes', None)
-        exclude_setinputsizes = kwargs.pop('exclude_setinputsizes', None)
+        auto_setinputsizes = kwargs.pop("auto_setinputsizes", None)
+        exclude_setinputsizes = kwargs.pop("exclude_setinputsizes", None)
         if auto_setinputsizes or exclude_setinputsizes:
             util.warn_deprecated(
                 "auto_setinputsizes and exclude_setinputsizes are deprecated. "
                 "Modern cx_Oracle only requires that LOB types are part "
                 "of this behavior, and these parameters no longer have any "
-                "effect.")
-        allow_twophase = kwargs.pop('allow_twophase', None)
+                "effect."
+            )
+        allow_twophase = kwargs.pop("allow_twophase", None)
         if allow_twophase is not None:
             util.warn.deprecated(
                 "allow_twophase is deprecated.  The cx_Oracle dialect no "
@@ -645,18 +830,16 @@ class OracleDialect_cx_oracle(OracleDialect):
             )
 
     def _parse_cx_oracle_ver(self, version):
-        m = re.match(r'(\d+)\.(\d+)(?:\.(\d+))?', version)
+        m = re.match(r"(\d+)\.(\d+)(?:\.(\d+))?", version)
         if m:
-            return tuple(
-                int(x)
-                for x in m.group(1, 2, 3)
-                if x is not None)
+            return tuple(int(x) for x in m.group(1, 2, 3) if x is not None)
         else:
             return (0, 0, 0)
 
     @classmethod
     def dbapi(cls):
         import cx_Oracle
+
         return cx_Oracle
 
     def initialize(self, connection):
@@ -675,15 +858,18 @@ class OracleDialect_cx_oracle(OracleDialect):
 
         self._decimal_char = connection.scalar(
             "select value from nls_session_parameters "
-            "where parameter = 'NLS_NUMERIC_CHARACTERS'")[0]
-        if self._decimal_char != '.':
+            "where parameter = 'NLS_NUMERIC_CHARACTERS'"
+        )[0]
+        if self._decimal_char != ".":
             _detect_decimal = self._detect_decimal
             _to_decimal = self._to_decimal
 
             self._detect_decimal = lambda value: _detect_decimal(
-                value.replace(self._decimal_char, "."))
+                value.replace(self._decimal_char, ".")
+            )
             self._to_decimal = lambda value: _to_decimal(
-                value.replace(self._decimal_char, "."))
+                value.replace(self._decimal_char, ".")
+            )
 
     def _detect_decimal(self, value):
         if "." in value:
@@ -702,13 +888,16 @@ class OracleDialect_cx_oracle(OracleDialect):
         dialect = self
         cx_Oracle = dialect.dbapi
 
-        number_handler = _OracleNUMBER(asdecimal=True).\
-            _cx_oracle_outputtypehandler(dialect)
-        float_handler = _OracleNUMBER(asdecimal=False).\
-            _cx_oracle_outputtypehandler(dialect)
+        number_handler = _OracleNUMBER(
+            asdecimal=True
+        )._cx_oracle_outputtypehandler(dialect)
+        float_handler = _OracleNUMBER(
+            asdecimal=False
+        )._cx_oracle_outputtypehandler(dialect)
 
-        def output_type_handler(cursor, name, default_type,
-                                size, precision, scale):
+        def output_type_handler(
+            cursor, name, default_type, size, precision, scale
+        ):
             if default_type == cx_Oracle.NUMBER:
                 if not dialect.coerce_to_decimal:
                     return None
@@ -719,7 +908,8 @@ class OracleDialect_cx_oracle(OracleDialect):
                         cx_Oracle.STRING,
                         255,
                         outconverter=dialect._detect_decimal,
-                        arraysize=cursor.arraysize)
+                        arraysize=cursor.arraysize,
+                    )
                 elif precision and scale > 0:
                     return number_handler(
                         cursor, name, default_type, size, precision, scale
@@ -730,18 +920,23 @@ class OracleDialect_cx_oracle(OracleDialect):
                     )
 
             # allow all strings to come back natively as Unicode
-            elif dialect.coerce_to_unicode and \
-                    default_type in (cx_Oracle.STRING, cx_Oracle.FIXED_CHAR):
-                return cursor.var(
-                    util.text_type, size, cursor.arraysize
-                )
+            elif dialect.coerce_to_unicode and default_type in (
+                cx_Oracle.STRING,
+                cx_Oracle.FIXED_CHAR,
+            ):
+                return cursor.var(util.text_type, size, cursor.arraysize)
             elif dialect.auto_convert_lobs and default_type in (
-                    cx_Oracle.CLOB, cx_Oracle.NCLOB, cx_Oracle.BLOB
+                cx_Oracle.CLOB,
+                cx_Oracle.NCLOB,
+                cx_Oracle.BLOB,
             ):
                 return cursor.var(
-                    default_type, size, cursor.arraysize,
-                    outconverter=lambda value: value.read()
+                    default_type,
+                    size,
+                    cursor.arraysize,
+                    outconverter=lambda value: value.read(),
                 )
+
         return output_type_handler
 
     def on_connect(self):
@@ -756,14 +951,19 @@ class OracleDialect_cx_oracle(OracleDialect):
     def create_connect_args(self, url):
         dialect_opts = dict(url.query)
 
-        for opt in ('use_ansi', 'auto_setinputsizes', 'auto_convert_lobs',
-                    'threaded', 'allow_twophase'):
+        for opt in (
+            "use_ansi",
+            "auto_setinputsizes",
+            "auto_convert_lobs",
+            "threaded",
+            "allow_twophase",
+        ):
             if opt in dialect_opts:
                 util.coerce_kw_type(dialect_opts, opt, bool)
                 setattr(self, opt, dialect_opts[opt])
 
         database = url.database
-        service_name = dialect_opts.get('service_name', None)
+        service_name = dialect_opts.get("service_name", None)
         if database or service_name:
             # if we have a database, then we have a remote host
             port = url.port
@@ -775,55 +975,50 @@ class OracleDialect_cx_oracle(OracleDialect):
             if database and service_name:
                 raise exc.InvalidRequestError(
                     '"service_name" option shouldn\'t '
-                    'be used with a "database" part of the url')
+                    'be used with a "database" part of the url'
+                )
             if database:
-                makedsn_kwargs = {'sid': database}
+                makedsn_kwargs = {"sid": database}
             if service_name:
-                makedsn_kwargs = {'service_name': service_name}
+                makedsn_kwargs = {"service_name": service_name}
 
             dsn = self.dbapi.makedsn(url.host, port, **makedsn_kwargs)
         else:
             # we have a local tnsname
             dsn = url.host
 
-        opts = dict(
-            threaded=self.threaded,
-        )
+        opts = dict(threaded=self.threaded)
 
         if dsn is not None:
-            opts['dsn'] = dsn
+            opts["dsn"] = dsn
         if url.password is not None:
-            opts['password'] = url.password
+            opts["password"] = url.password
         if url.username is not None:
-            opts['user'] = url.username
+            opts["user"] = url.username
 
-        if 'mode' in url.query:
-            opts['mode'] = url.query['mode']
-            if isinstance(opts['mode'], util.string_types):
-                mode = opts['mode'].upper()
-                if mode == 'SYSDBA':
-                    opts['mode'] = self.dbapi.SYSDBA
-                elif mode == 'SYSOPER':
-                    opts['mode'] = self.dbapi.SYSOPER
+        if "mode" in url.query:
+            opts["mode"] = url.query["mode"]
+            if isinstance(opts["mode"], util.string_types):
+                mode = opts["mode"].upper()
+                if mode == "SYSDBA":
+                    opts["mode"] = self.dbapi.SYSDBA
+                elif mode == "SYSOPER":
+                    opts["mode"] = self.dbapi.SYSOPER
                 else:
-                    util.coerce_kw_type(opts, 'mode', int)
+                    util.coerce_kw_type(opts, "mode", int)
         return ([], opts)
 
     def _get_server_version_info(self, connection):
-        return tuple(
-            int(x)
-            for x in connection.connection.version.split('.')
-        )
+        return tuple(int(x) for x in connection.connection.version.split("."))
 
     def is_disconnect(self, e, connection, cursor):
         error, = e.args
         if isinstance(
-                e,
-                (self.dbapi.InterfaceError, self.dbapi.DatabaseError)
+            e, (self.dbapi.InterfaceError, self.dbapi.DatabaseError)
         ) and "not connected" in str(e):
             return True
 
-        if hasattr(error, 'code'):
+        if hasattr(error, "code"):
             # ORA-00028: your session has been killed
             # ORA-03114: not connected to ORACLE
             # ORA-03113: end-of-file on communication channel
@@ -846,8 +1041,8 @@ class OracleDialect_cx_oracle(OracleDialect):
 
         """
 
-        id = random.randint(0, 2 ** 128)
-        return (0x1234, "%032x" % id, "%032x" % 9)
+        id_ = random.randint(0, 2 ** 128)
+        return (0x1234, "%032x" % id_, "%032x" % 9)
 
     def do_executemany(self, cursor, statement, parameters, context=None):
         if isinstance(parameters, tuple):
@@ -859,22 +1054,25 @@ class OracleDialect_cx_oracle(OracleDialect):
 
     def do_prepare_twophase(self, connection, xid):
         result = connection.connection.prepare()
-        connection.info['cx_oracle_prepared'] = result
+        connection.info["cx_oracle_prepared"] = result
 
-    def do_rollback_twophase(self, connection, xid, is_prepared=True,
-                             recover=False):
+    def do_rollback_twophase(
+        self, connection, xid, is_prepared=True, recover=False
+    ):
         self.do_rollback(connection.connection)
 
-    def do_commit_twophase(self, connection, xid, is_prepared=True,
-                           recover=False):
+    def do_commit_twophase(
+        self, connection, xid, is_prepared=True, recover=False
+    ):
         if not is_prepared:
             self.do_commit(connection.connection)
         else:
-            oci_prepared = connection.info['cx_oracle_prepared']
+            oci_prepared = connection.info["cx_oracle_prepared"]
             if oci_prepared:
                 self.do_commit(connection.connection)
 
     def do_recover_twophase(self, connection):
-        connection.info.pop('cx_oracle_prepared', None)
+        connection.info.pop("cx_oracle_prepared", None)
+
 
 dialect = OracleDialect_cx_oracle
