@@ -1,5 +1,5 @@
 # sql/types_api.py
-# Copyright (C) 2005-2016 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -10,10 +10,12 @@
 """
 
 
-from .. import exc, util
 from . import operators
-from .visitors import Visitable, VisitableType
 from .base import SchemaEventTarget
+from .visitors import Visitable
+from .visitors import VisitableType
+from .. import exc
+from .. import util
 
 # these are back-assigned by sqltypes.
 BOOLEANTYPE = None
@@ -49,7 +51,8 @@ class TypeEngine(Visitable):
 
 
         """
-        __slots__ = 'expr', 'type'
+
+        __slots__ = "expr", "type"
 
         default_comparator = None
 
@@ -57,16 +60,15 @@ class TypeEngine(Visitable):
             self.expr = expr
             self.type = expr.type
 
-        @util.dependencies('sqlalchemy.sql.default_comparator')
+        @util.dependencies("sqlalchemy.sql.default_comparator")
         def operate(self, default_comparator, op, *other, **kwargs):
             o = default_comparator.operator_lookup[op.__name__]
             return o[0](self.expr, op, *(other + o[1:]), **kwargs)
 
-        @util.dependencies('sqlalchemy.sql.default_comparator')
+        @util.dependencies("sqlalchemy.sql.default_comparator")
         def reverse_operate(self, default_comparator, op, other, **kwargs):
             o = default_comparator.operator_lookup[op.__name__]
-            return o[0](self.expr, op, other,
-                        reverse=True, *o[1:], **kwargs)
+            return o[0](self.expr, op, other, reverse=True, *o[1:], **kwargs)
 
         def _adapt_expression(self, op, other_comparator):
             """evaluate the return type of <self> <op> <othertype>,
@@ -93,10 +95,11 @@ class TypeEngine(Visitable):
             boolean comparison or special SQL keywords like MATCH or BETWEEN.
 
             """
+
             return op, self.type
 
         def __reduce__(self):
-            return _reconstitute_comparator, (self.expr, )
+            return _reconstitute_comparator, (self.expr,)
 
     hashable = True
     """Flag, if False, means values from this type aren't hashable.
@@ -182,6 +185,13 @@ class TypeEngine(Visitable):
         the :obj:`~.expression.null` SQL construct in an INSERT statement
         or associated with an ORM-mapped attribute.
 
+        .. note::
+
+            The "evaulates none" flag does **not** apply to a value
+            of ``None`` passed to :paramref:`.Column.default` or
+            :paramref:`.Column.server_default`; in these cases, ``None``
+            still means "no default".
+
         .. versionadded:: 1.1
 
         .. seealso::
@@ -305,8 +315,10 @@ class TypeEngine(Visitable):
 
         """
 
-        return self.__class__.column_expression.__code__ \
+        return (
+            self.__class__.column_expression.__code__
             is not TypeEngine.column_expression.__code__
+        )
 
     def bind_expression(self, bindvalue):
         """"Given a bind value (i.e. a :class:`.BindParameter` instance),
@@ -343,8 +355,14 @@ class TypeEngine(Visitable):
 
         """
 
-        return self.__class__.bind_expression.__code__ \
+        return (
+            self.__class__.bind_expression.__code__
             is not TypeEngine.bind_expression.__code__
+        )
+
+    @staticmethod
+    def _to_instance(cls_or_self):
+        return to_instance(cls_or_self)
 
     def compare_values(self, x, y):
         """Compare two values for equality."""
@@ -380,7 +398,7 @@ class TypeEngine(Visitable):
         raise NotImplementedError()
 
     def with_variant(self, type_, dialect_name):
-        """Produce a new type object that will utilize the given
+        r"""Produce a new type object that will utilize the given
         type when applied to the dialect of the given name.
 
         e.g.::
@@ -398,7 +416,7 @@ class TypeEngine(Visitable):
         itself provides a :meth:`.Variant.with_variant`
         that can be called repeatedly.
 
-        :param type_: a :class:`.TypeEngine` that will be selected
+        :param type\_: a :class:`.TypeEngine` that will be selected
          as a variant from the originating type, when a dialect
          of the given name is in use.
         :param dialect_name: base name of the dialect which uses
@@ -429,27 +447,38 @@ class TypeEngine(Visitable):
 
         """
         try:
-            return dialect._type_memos[self]['impl']
+            return dialect._type_memos[self]["impl"]
         except KeyError:
-            return self._dialect_info(dialect)['impl']
+            return self._dialect_info(dialect)["impl"]
+
+    def _unwrapped_dialect_impl(self, dialect):
+        """Return the 'unwrapped' dialect impl for this type.
+
+        For a type that applies wrapping logic (e.g. TypeDecorator), give
+        us the real, actual dialect-level type that is used.
+
+        This is used for class-based lookups by dialects.
+
+        """
+        return self.dialect_impl(dialect)
 
     def _cached_literal_processor(self, dialect):
         """Return a dialect-specific literal processor for this type."""
         try:
-            return dialect._type_memos[self]['literal']
+            return dialect._type_memos[self]["literal"]
         except KeyError:
             d = self._dialect_info(dialect)
-            d['literal'] = lp = d['impl'].literal_processor(dialect)
+            d["literal"] = lp = d["impl"].literal_processor(dialect)
             return lp
 
     def _cached_bind_processor(self, dialect):
         """Return a dialect-specific bind processor for this type."""
 
         try:
-            return dialect._type_memos[self]['bind']
+            return dialect._type_memos[self]["bind"]
         except KeyError:
             d = self._dialect_info(dialect)
-            d['bind'] = bp = d['impl'].bind_processor(dialect)
+            d["bind"] = bp = d["impl"].bind_processor(dialect)
             return bp
 
     def _cached_result_processor(self, dialect, coltype):
@@ -462,9 +491,18 @@ class TypeEngine(Visitable):
             # key assumption: DBAPI type codes are
             # constants.  Else this dictionary would
             # grow unbounded.
-            d[coltype] = rp = d['impl'].result_processor(dialect, coltype)
+            d[coltype] = rp = d["impl"].result_processor(dialect, coltype)
             return rp
 
+    def _cached_custom_processor(self, dialect, key, fn):
+        try:
+            return dialect._type_memos[self][key]
+        except KeyError:
+            d = self._dialect_info(dialect)
+            impl = d["impl"]
+            d[key] = result = fn(impl)
+            return result
+
     def _dialect_info(self, dialect):
         """Return a dialect-specific registry which
         caches a dialect-specific implementation, bind processing
@@ -478,7 +516,7 @@ class TypeEngine(Visitable):
                 impl = self.adapt(type(self))
             # this can't be self, else we create a cycle
             assert impl is not self
-            dialect._type_memos[self] = d = {'impl': impl}
+            dialect._type_memos[self] = d = {"impl": impl}
             return d
 
     def _gen_dialect_impl(self, dialect):
@@ -514,8 +552,10 @@ class TypeEngine(Visitable):
 
         """
         _coerced_type = _resolve_value_to_type(value)
-        if _coerced_type is NULLTYPE or _coerced_type._type_affinity \
-                is self._type_affinity:
+        if (
+            _coerced_type is NULLTYPE
+            or _coerced_type._type_affinity is self._type_affinity
+        ):
             return self
         else:
             return _coerced_type
@@ -551,8 +591,9 @@ class TypeEngine(Visitable):
 
     def __str__(self):
         if util.py2k:
-            return unicode(self.compile()).\
-                encode('ascii', 'backslashreplace')
+            return unicode(self.compile()).encode(  # noqa
+                "ascii", "backslashreplace"
+            )  # noqa
         else:
             return str(self.compile())
 
@@ -610,15 +651,16 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
        ``type_expression``, if it receives ``**kw`` in its signature.
 
     """
+
     __visit_name__ = "user_defined"
 
-    ensure_kwarg = 'get_col_spec'
+    ensure_kwarg = "get_col_spec"
 
     class Comparator(TypeEngine.Comparator):
         __slots__ = ()
 
         def _adapt_expression(self, op, other_comparator):
-            if hasattr(self.type, 'adapt_operator'):
+            if hasattr(self.type, "adapt_operator"):
                 util.warn_deprecated(
                     "UserDefinedType.adapt_operator is deprecated.  Create "
                     "a UserDefinedType.Comparator subclass instead which "
@@ -627,7 +669,9 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
                 )
                 return self.type.adapt_operator(op), self.type
             else:
-                return op, self.type
+                return super(
+                    UserDefinedType.Comparator, self
+                )._adapt_expression(op, other_comparator)
 
     comparator_factory = Comparator
 
@@ -650,6 +694,73 @@ class UserDefinedType(util.with_metaclass(VisitableCheckKWArg, TypeEngine)):
         return self
 
 
+class Emulated(object):
+    """Mixin for base types that emulate the behavior of a DB-native type.
+
+    An :class:`.Emulated` type will use an available database type
+    in conjunction with Python-side routines and/or database constraints
+    in order to approximate the behavior of a database type that is provided
+    natively by some backends.  When a native-providing backend is in
+    use, the native version of the type is used.  This native version
+    should include the :class:`.NativeForEmulated` mixin to allow it to be
+    distinguished from :class:`.Emulated`.
+
+    Current examples of :class:`.Emulated` are:  :class:`.Interval`,
+    :class:`.Enum`, :class:`.Boolean`.
+
+    .. versionadded:: 1.2.0b3
+
+    """
+
+    def adapt_to_emulated(self, impltype, **kw):
+        """Given an impl class, adapt this type to the impl assuming "emulated".
+
+        The impl should also be an "emulated" version of this type,
+        most likely the same class as this type itself.
+
+        e.g.: sqltypes.Enum adapts to the Enum class.
+
+        """
+        return super(Emulated, self).adapt(impltype, **kw)
+
+    def adapt(self, impltype, **kw):
+        if hasattr(impltype, "adapt_emulated_to_native"):
+
+            if self.native:
+                # native support requested, dialect gave us a native
+                # implementor, pass control over to it
+                return impltype.adapt_emulated_to_native(self, **kw)
+            else:
+                # impltype adapts to native, and we are not native,
+                # so reject the impltype in favor of "us"
+                impltype = self.__class__
+
+        if issubclass(impltype, self.__class__):
+            return self.adapt_to_emulated(impltype, **kw)
+        else:
+            return super(Emulated, self).adapt(impltype, **kw)
+
+
+class NativeForEmulated(object):
+    """Indicates DB-native types supported by an :class:`.Emulated` type.
+
+    .. versionadded:: 1.2.0b3
+
+    """
+
+    @classmethod
+    def adapt_emulated_to_native(cls, impl, **kw):
+        """Given an impl, adapt this type's class to the impl assuming "native".
+
+        The impl will be an :class:`.Emulated` class but not a
+        :class:`.NativeForEmulated`.
+
+        e.g.: postgresql.ENUM produces a type given an Enum instance.
+
+        """
+        return cls(**kw)
+
+
 class TypeDecorator(SchemaEventTarget, TypeEngine):
     """Allows the creation of types which add additional functionality
     to an existing type.
@@ -771,14 +882,16 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        if not hasattr(self.__class__, 'impl'):
-            raise AssertionError("TypeDecorator implementations "
-                                 "require a class-level variable "
-                                 "'impl' which refers to the class of "
-                                 "type being decorated")
+        if not hasattr(self.__class__, "impl"):
+            raise AssertionError(
+                "TypeDecorator implementations "
+                "require a class-level variable "
+                "'impl' which refers to the class of "
+                "type being decorated"
+            )
         self.impl = to_instance(self.__class__.impl, *args, **kwargs)
 
-    coerce_to_is_types = (util.NoneType, )
+    coerce_to_is_types = (util.NoneType,)
     """Specify those Python types which should be coerced at the expression
     level to "IS <constant>" when compared using ``==`` (and same for
     ``IS NOT`` in conjunction with ``!=``.
@@ -803,24 +916,27 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         __slots__ = ()
 
         def operate(self, op, *other, **kwargs):
-            kwargs['_python_is_types'] = self.expr.type.coerce_to_is_types
+            kwargs["_python_is_types"] = self.expr.type.coerce_to_is_types
             return super(TypeDecorator.Comparator, self).operate(
-                op, *other, **kwargs)
+                op, *other, **kwargs
+            )
 
         def reverse_operate(self, op, other, **kwargs):
-            kwargs['_python_is_types'] = self.expr.type.coerce_to_is_types
+            kwargs["_python_is_types"] = self.expr.type.coerce_to_is_types
             return super(TypeDecorator.Comparator, self).reverse_operate(
-                op, other, **kwargs)
+                op, other, **kwargs
+            )
 
     @property
     def comparator_factory(self):
         if TypeDecorator.Comparator in self.impl.comparator_factory.__mro__:
             return self.impl.comparator_factory
         else:
-            return type("TDComparator",
-                        (TypeDecorator.Comparator,
-                         self.impl.comparator_factory),
-                        {})
+            return type(
+                "TDComparator",
+                (TypeDecorator.Comparator, self.impl.comparator_factory),
+                {},
+            )
 
     def _gen_dialect_impl(self, dialect):
         """
@@ -833,13 +949,14 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         # otherwise adapt the impl type, link
         # to a copy of this TypeDecorator and return
         # that.
-        typedesc = self.load_dialect_impl(dialect).dialect_impl(dialect)
+        typedesc = self._unwrapped_dialect_impl(dialect)
         tt = self.copy()
         if not isinstance(tt, self.__class__):
-            raise AssertionError('Type object %s does not properly '
-                                 'implement the copy() method, it must '
-                                 'return an object of type %s' %
-                                 (self, self.__class__))
+            raise AssertionError(
+                "Type object %s does not properly "
+                "implement the copy() method, it must "
+                "return an object of type %s" % (self, self.__class__)
+            )
         tt.impl = typedesc
         return tt
 
@@ -851,13 +968,17 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         return self.impl._type_affinity
 
     def _set_parent(self, column):
-        """Support SchemaEentTarget"""
+        """Support SchemaEventTarget"""
+
+        super(TypeDecorator, self)._set_parent(column)
 
         if isinstance(self.impl, SchemaEventTarget):
             self.impl._set_parent(column)
 
     def _set_parent_with_dispatch(self, parent):
-        """Support SchemaEentTarget"""
+        """Support SchemaEventTarget"""
+
+        super(TypeDecorator, self)._set_parent_with_dispatch(parent)
 
         if isinstance(self.impl, SchemaEventTarget):
             self.impl._set_parent_with_dispatch(parent)
@@ -896,6 +1017,17 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         """
         return self.impl
 
+    def _unwrapped_dialect_impl(self, dialect):
+        """Return the 'unwrapped' dialect impl for this type.
+
+        For a type that applies wrapping logic (e.g. TypeDecorator), give
+        us the real, actual dialect-level type that is used.
+
+        This is used for class-based lookups by dialects.
+
+        """
+        return self.load_dialect_impl(dialect).dialect_impl(dialect)
+
     def __getattr__(self, key):
         """Proxy all other undefined accessors to the underlying
         implementation."""
@@ -978,8 +1110,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        return self.__class__.process_bind_param.__code__ \
+        return (
+            self.__class__.process_bind_param.__code__
             is not TypeDecorator.process_bind_param.__code__
+        )
 
     @util.memoized_property
     def _has_literal_processor(self):
@@ -988,8 +1122,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
 
         """
 
-        return self.__class__.process_literal_param.__code__ \
+        return (
+            self.__class__.process_literal_param.__code__
             is not TypeDecorator.process_literal_param.__code__
+        )
 
     def literal_processor(self, dialect):
         """Provide a literal processing function for the given
@@ -1026,9 +1162,12 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         if process_param:
             impl_processor = self.impl.literal_processor(dialect)
             if impl_processor:
+
                 def process(value):
                     return impl_processor(process_param(value, dialect))
+
             else:
+
                 def process(value):
                     return process_param(value, dialect)
 
@@ -1059,10 +1198,12 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
             process_param = self.process_bind_param
             impl_processor = self.impl.bind_processor(dialect)
             if impl_processor:
+
                 def process(value):
                     return impl_processor(process_param(value, dialect))
 
             else:
+
                 def process(value):
                     return process_param(value, dialect)
 
@@ -1079,8 +1220,10 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         exception throw.
 
         """
-        return self.__class__.process_result_value.__code__ \
+        return (
+            self.__class__.process_result_value.__code__
             is not TypeDecorator.process_result_value.__code__
+        )
 
     def result_processor(self, dialect, coltype):
         """Provide a result value processing function for the given
@@ -1104,13 +1247,14 @@ class TypeDecorator(SchemaEventTarget, TypeEngine):
         """
         if self._has_result_processor:
             process_value = self.process_result_value
-            impl_processor = self.impl.result_processor(dialect,
-                                                        coltype)
+            impl_processor = self.impl.result_processor(dialect, coltype)
             if impl_processor:
+
                 def process(value):
                     return process_value(impl_processor(value), dialect)
 
             else:
+
                 def process(value):
                     return process_value(value, dialect)
 
@@ -1202,18 +1346,43 @@ class Variant(TypeDecorator):
         self.impl = base
         self.mapping = mapping
 
+    def coerce_compared_value(self, operator, value):
+        result = self.impl.coerce_compared_value(operator, value)
+        if result is self.impl:
+            return self
+        else:
+            return result
+
     def load_dialect_impl(self, dialect):
         if dialect.name in self.mapping:
             return self.mapping[dialect.name]
         else:
             return self.impl
 
+    def _set_parent(self, column):
+        """Support SchemaEventTarget"""
+
+        if isinstance(self.impl, SchemaEventTarget):
+            self.impl._set_parent(column)
+        for impl in self.mapping.values():
+            if isinstance(impl, SchemaEventTarget):
+                impl._set_parent(column)
+
+    def _set_parent_with_dispatch(self, parent):
+        """Support SchemaEventTarget"""
+
+        if isinstance(self.impl, SchemaEventTarget):
+            self.impl._set_parent_with_dispatch(parent)
+        for impl in self.mapping.values():
+            if isinstance(impl, SchemaEventTarget):
+                impl._set_parent_with_dispatch(parent)
+
     def with_variant(self, type_, dialect_name):
-        """Return a new :class:`.Variant` which adds the given
+        r"""Return a new :class:`.Variant` which adds the given
         type + dialect name to the mapping, in addition to the
         mapping present in this :class:`.Variant`.
 
-        :param type_: a :class:`.TypeEngine` that will be selected
+        :param type\_: a :class:`.TypeEngine` that will be selected
          as a variant from the originating type, when a dialect
          of the given name is in use.
         :param dialect_name: base name of the dialect which uses
@@ -1224,7 +1393,8 @@ class Variant(TypeDecorator):
         if dialect_name in self.mapping:
             raise exc.ArgumentError(
                 "Dialect '%s' is already present in "
-                "the mapping for this Variant" % dialect_name)
+                "the mapping for this Variant" % dialect_name
+            )
         mapping = self.mapping.copy()
         mapping[dialect_name] = type_
         return Variant(self.impl, mapping)
@@ -1266,6 +1436,6 @@ def adapt_type(typeobj, colspecs):
     # but it turns out the originally given "generic" type
     # is actually a subclass of our resulting type, then we were already
     # given a more specific type than that required; so use that.
-    if (issubclass(typeobj.__class__, impltype)):
+    if issubclass(typeobj.__class__, impltype):
         return typeobj
     return typeobj.adapt(impltype)
